// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the GNU General Public License version 2 ("the License").
// You may not use this file except in compliance with the License.

// FIXME
package org.ibex.core;

// FIXME: are traps on x/y meaningful?
// FIXME: if we trap on cols, then set rows to 0 (forcing cols to 1), does the cols trap get triggered?
// FIXME: if we change min{width/height}, thereby forcing a change to max{min/height}, does a trap on those get triggered?
// FIXME: trap on numchildren?  replaces ChildChanged?
// FIXME: trap on visible, trigger when parent visibility changes

// FIXME: ax/ay nonsense
// FIXME: mouse move/release still needs to propagate to boxen in which the mouse was pressed and is still held down

// FEATURE: mark to reflow starting with a certain child
// FEATURE: reintroduce surface.abort

import java.util.*;
import org.ibex.js.*;
import org.ibex.util.*;
import org.ibex.graphics.*;

/**
 *  <p>
 *  Encapsulates the data for a single Ibex box as well as all layout
 *  rendering logic.
 *  </p>
 *
 *  <p>The rendering process consists of four phases; each requires
 *     one DFS pass over the tree</p>
 *  <ol><li> <b>pack()</b>: each box sets its childrens' row/col
 *  <ol><li> <b>constrain()</b>: contentwidth is computed
 *      <li> <b>resize()</b>: width/height and x/y positions are set
 *      <li> <b>render()</b>: children draw their content onto the PixelBuffer.
 *  </ol>
 *
 *  The first three passes together are called the <i>reflow</i>
 *  phase.  Reflowing is done in a seperate pass since SizeChanges
 *  trigger a Surface.abort; if rendering were done in the same pass,
 *  rendering work done prior to the Surface.abort would be wasted.
 */
public final class Box extends JS.Obj implements Callable {

    private static final JS.Method METHOD = new JS.Method();

    // Macros //////////////////////////////////////////////////////////////////////

    final void REPLACE() { for(Box b2 = this; b2 != null && !b2.test(REPLACE); b2 = b2.parent) b2.set(REPLACE); }
    final void RECONSTRAIN() { for(Box b2 = this; b2 != null && !b2.test(RECONSTRAIN); b2 = b2.parent) b2.set(RECONSTRAIN); }
    final void REPACK() { for(Box b2 = this; b2 != null && !b2.test(REPACK); b2 = b2.parent) b2.set(REPACK); }

    //#define CHECKSET_SHORT(prop) short nu = (short)JSU.toInt(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_INT(prop) int nu = JSU.toInt(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_FLAG(flag) boolean nu = JSU.toBoolean(value); if (nu == test(flag)) break; if (nu) set(flag); else clear(flag);
    //#define CHECKSET_BOOLEAJSU.N(prop) boolean nu = JSU.toBoolean(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_STRING(prop) if ((value==null&&prop==null)||(value!=null&&JSU.toString(value).equals(prop))) break; prop=JSU.toString(value);

    // FIXME memory leak
    static Basket.Map boxToCursor = new Basket.HashMap(500, 3);

    static final Font DEFAULT_FONT;
    static {
        try { DEFAULT_FONT = Font.getFont((JS)Main.builtin.get(JSU.S("fonts/vera/Vera.ttf")), 10); }
        catch(JSExn e) { throw new Error("Error loading default font: " + e); }
    }


    // Flags //////////////////////////////////////////////////////////////////////

    static final int MOUSEINSIDE  = 0x00000001;
    static final int VISIBLE      = 0x00000002;
    static final int PACKED       = 0x00000004;
    public static final int HSHRINK      = 0x00000008;
    public static final int VSHRINK      = 0x00000010;
    static final int BLACK        = 0x00000020;  // for red-black code

    static final int FIXED        = 0x00000040;
    static final boolean ROWS     = true;
    static final boolean COLS     = false;

    static final int ISROOT       = 0x00000080;
    static final int REPACK       = 0x00000100;
    static final int RECONSTRAIN  = 0x00000200;
    static final int REPLACE      = 0x00000400;

    static final int ALIGN_TOP    = 0x00001000;
    static final int ALIGN_BOTTOM = 0x00002000;
    static final int ALIGN_LEFT   = 0x00004000;
    static final int ALIGN_RIGHT  = 0x00008000;
    static final int ALIGNS       = 0x0000f000;
    static final int CURSOR       = 0x00010000;  // if true, this box has a cursor in the cursor hash; FEATURE: GC issues?
    static final int CLIP         = 0x00020000;
    static final int STOP_UPWARD_PROPAGATION    = 0x00040000;
    static final int MOVED         = 0x00080000;


    // Instance Data //////////////////////////////////////////////////////////////////////

    Box parent = null;
    Box redirect = this;
    int flags = VISIBLE | PACKED | REPACK | RECONSTRAIN | REPLACE | FIXED | STOP_UPWARD_PROPAGATION | CLIP | MOVED;

    private BalancedTree bt;
    
    private String text = null;
    private Font font = DEFAULT_FONT; 
    private Picture texture = null;
    private short strokewidth = 1;
    public int fillcolor = 0x00000000;
    private int strokecolor = 0xFF000000;

    private int aspect = 0;

    // specified directly by user
    public int minwidth = 0;
    public int maxwidth = Integer.MAX_VALUE;
    public int minheight = 0;
    public int maxheight = Integer.MAX_VALUE;
    private short rows = 1;
    private short cols = 0;
    private short rowspan = 1;
    private short colspan = 1;

    // computed during reflow
    private short row = 0;
    private short col = 0;
    public int x = 0;
    public int y = 0;
    public int ax = 0;   // FEATURE: roll these into x/y; requires lots of changes
    public int ay = 0;   // FEATURE: roll these into x/y; requires lots of changes; perhaps y()?
    public int width = 0;
    public int height = 0;
    public int contentwidth = 0;      // == max(minwidth, textwidth, sum(child.contentwidth))
    public int contentheight = 0;

    private Path path = null;
    /*
    private Affine transform = null;
    private VectorGraphics.RasterPath rpath = null;
    private Affine rtransform = null;
    */

    // Instance Methods /////////////////////////////////////////////////////////////////////

    public final int fontSize() { return font == null ? DEFAULT_FONT.pointsize : font.pointsize; }

    /** invoked when a resource needed to render ourselves finishes loading */
    public Object run(Object o) throws JSExn {
        if (texture == null) { Log.warn(Box.class, "perform() called with null texture"); return null; }
        if (texture.isLoaded) {
            setWidth(max(texture.width, minwidth), maxwidth); 
            setHeight(max(texture.height, minheight), maxheight); 
            dirty(); }
        else { JS res = texture.stream; texture = null; throw new JSExn("image not found: "+res.unclone()); }
        return null;
    }

    // FEATURE: use cx2/cy2 format
    /** Adds the intersection of (x,y,w,h) and the node's current actual geometry to the Surface's dirty list */
    public void dirty() { dirty(0, 0, width, height); }
    public void dirty(int x, int y, int w, int h) {
        for(Box cur = this; cur != null; cur = cur.parent) {
            // x and y have a different meaning on the root box
            if (cur.parent != null && cur.test(CLIP)) {
                w = min(x + w, cur.width) - max(x, 0);
                h = min(y + h, cur.height) - max(y, 0);
                x = max(x, 0);
                y = max(y, 0);
            }
            if (w <= 0 || h <= 0) return;
            if (cur.parent == null && cur.getSurface() != null) cur.getSurface().dirty(x, y, w, h);
            x += cur.x;
            y += cur.y;
        }
    }


    // Reflow ////////////////////////////////////////////////////////////////////////////////////////

    /** should only be invoked on the root box */
    public void reflow() {
        pack();
        resize(x, y,
               test(HSHRINK) ? contentwidth : maxwidth,
               test(VSHRINK) ? contentheight : maxheight);
        place();
    }
    
    private static Box[] frontier = new Box[65535];
    /** pack the boxes into rows and columns, compute contentwidth */
    public void pack() {
        if (!test(REPACK)) { constrain(); return; }
        boolean haskid = false;
        for(Box child = getChild(0); child != null; child = child.nextSibling()) { haskid = true; child.pack(); }
        if (!haskid) { clear(REPACK); constrain(); return; }
        int frontier_size = 0;
        //#repeat COLS/ROWS rows/cols cols/rows col/row row/col colspan/rowspan rowspan/colspan \
        //        contentheight/contentwidth contentwidth/contentheight
        if (test(FIXED) == COLS) {
            rows = 0;
            for(Box child = getChild(0); child != null; child = child.nextSibling()) {
                if (!child.test(PACKED) || !child.test(VISIBLE)) continue;
                if (cols == 1) { child.row = rows; rows += child.rowspan; child.col = 0; continue; }
                child.col = (short)(frontier_size <= 0 ? 0 : (frontier[frontier_size-1].col + frontier[frontier_size-1].colspan));
                child.row = (short)(frontier_size <= 0 ? 0 : frontier[frontier_size-1].row);
                if (child.col + min(cols,child.colspan) > cols) { child.col = 0; child.row++; }
                for(int i=0; i<frontier_size; i++)
                    if (frontier[i].row + frontier[i].rowspan <= child.row) {
                        frontier[i--] = frontier[--frontier_size]; frontier[frontier_size] = null;
                    } else if (frontier[i].col<child.col+min(cols,child.colspan)&&frontier[i].col+frontier[i].colspan>child.col) {
                        child.col = (short)(frontier[i].col + frontier[i].colspan);
                        if (child.col + min(cols,child.colspan) > cols) {
                            child.row = (short)(frontier[i].row + frontier[i].rowspan);
                            for(i--; i>0; i--) child.row = (short)min(row, frontier[i].row + frontier[i].rowspan);
                            child.col = (short)0;
                        }
                        i = -1;
                    } else break;
                frontier[frontier_size++] = child;
            }
            for(int i=0; i<frontier_size; i++){ rows=(short)max(rows, frontier[i].row + frontier[i].rowspan); frontier[i] = null; }
        }
        //#end
        clear(REPACK);
        set(RECONSTRAIN);   // FIXME: be smarter / more incremental
        constrain();
    }

    public void constrain() {
        if (!test(RECONSTRAIN)) return;
        solve(true);
        //#repeat contentwidth/contentheight contentheight/contentwidth minwidth/minheight row/col col/row \
        //        textwidth/textheight maxwidth/maxheight cols/rows rows/cols colspan/rowspan rowspan/colspan
        contentwidth = bound(minwidth,
                             max(contentwidth, font == null || text == null ? 0 : font.textwidth(text)),
                             maxwidth);
        //#end
        set(REPLACE); // FIXME: be smarter / more incremental
    }
    
    private final static JS SIZECHANGE = JSU.S("SizeChange");
    
    void resize(int x, int y, int width, int height) {
        if (x == this.x && y == this.y && width == this.width && height == this.height) return;
        boolean sizechange = (this.width != width || this.height != height);
        try { sizechange = sizechange && getTrap(SIZECHANGE) != null; } catch (JSExn e) {}
        int thisx = parent == null ? 0 : this.x;
        int thisy = parent == null ? 0 : this.y;
        Box who = (parent == null ? this : parent);
        if (this.x != x || this.y != y) set(MOVED);
        if (texture == null && (text == null || text.equals("")) && !test(MOVED)) {
            if ((fillcolor & 0xff000000) != 0 || parent == null) {
                who.dirty(thisx+min(this.width,width), thisy, Math.abs(width-this.width), max(this.height, height));
                who.dirty(thisx, thisy+min(this.height,height), max(this.width, width), Math.abs(height-this.height));
            }
            this.width = width; this.height = height; this.x = x; this.y = y;
        } else {
            who.dirty(thisx, thisy, this.width, this.height);
            this.width = width; this.height = height; this.x = x; this.y = y;
            dirty();
        }
        if (sizechange) putAndTriggerTrapsAndCatchExceptions(SIZECHANGE, JSU.T);
    }

    private float targetColumnSize = (float)0.0;
    private float targetRowSize = (float)0.0;
    private static float[] sizes = new float[65535];
    private static float[] sizes_v = new float[65535];
    private static int[] regions = new int[65535];
    private static int[] regions_v = new int[65535];
    private static int numregions = 0;
    private static int numregions_v = 0;

    void solve(boolean findMinimum) {
        int numkids = 0; for(Box c = firstPackedChild(); c != null; c = c.nextPackedSibling()) numkids++;
        //#repeat col/row colspan/rowspan contentwidth/contentheight width/height HSHRINK/VSHRINK numregions/numregions_v \
        //        maxwidth/maxheight cols/rows minwidth/minheight regions/regions_v targetColumnSize/targetRowSize sizes/sizes_v \
        //        HSHRINK/VSHRINK
        if (numkids == 0) {
            if (findMinimum) contentwidth = 0;
            else targetColumnSize = 0;
        } else if (cols == 1) {
            if (findMinimum) {
                contentwidth = 0;
                for(Box c = firstPackedChild(); c != null; c = c.nextPackedSibling())
                    contentwidth = max(contentwidth, c.contentwidth);
            } else {
                targetColumnSize = width;
            }
        } else if (cols > 1) do {

            // FIXME: cache these?
            // compute regions
            numregions = 0;
            for(Box c = firstPackedChild(); c != null; c = c.nextPackedSibling()) {
                regions[numregions++] = c.col;
                regions[numregions++] = min(cols, c.col+c.colspan);
            }
            Vec.sortInts(regions, 0, numregions);
            int j = 0;
            int newnumregions = numregions;
            for(int i=1; i<numregions; i++) {
                if (regions[j] != regions[i]) j++;
                else newnumregions--;
                regions[j] = regions[i];
            }
            numregions = newnumregions;
            if (regions[numregions-1] == cols) numregions--;
            else regions[numregions] = cols;

            int target = findMinimum ? 0 : Math.max(width, contentwidth);
            // priority 0: (inviolable) honor minwidths
            // priority 1: sum of columns no greater than parent
            // priority 2: honor maxwidths
            // priority 3: equalize columns
            float targetColumnSize = target == 0 ? 0 : this.targetColumnSize;
            float last_columnsize = 0;
            float last_total = 0;
            float total;
            boolean first = true;
            while(true) {
                total = (float)0.0;
                for(int r=0; r<numregions; r++) total += (sizes[r] = (float)(targetColumnSize * (regions[r+1]-regions[r])));
                int minregion = 0;
                for(Box child = firstPackedChild(); child != null; child = child.nextPackedSibling())
                    for(int r=(child.col==0?0:minregion); r<numregions; r++) {
                        if (regions[r+1] < child.col) continue;
                        if (regions[r] >= min(child.col+child.colspan,cols)) { minregion = r; break; }
                        total -= sizes[r];
                        int child_maxwidth = child.test(HSHRINK)?child.contentwidth:child.maxwidth;
                        if (sizes[r] <= (float)(targetColumnSize*(regions[r+1]-regions[r])))
                            if ((child.colspan * targetColumnSize) > (child_maxwidth + (float)0.5))
                                sizes[r] = (float)Math.min(sizes[r], (regions[r+1]-regions[r])*(child_maxwidth/child.colspan));
                        if ((child.colspan * targetColumnSize) < (child.contentwidth - (float)0.5))
                            sizes[r] = (float)Math.max(sizes[r], (regions[r+1]-regions[r])*(child.contentwidth/child.colspan));
                        total += sizes[r];
                    }
                float save = targetColumnSize;
                if (Math.abs(total - target) <= (float)1.0) break;
                if (!first) {
                    if (Math.abs(total - last_total) <= (float)1.0) break;
                } else {
                    last_columnsize = ((total - target) / (float)cols) + targetColumnSize;
                }
                if (total < target)      targetColumnSize += Math.abs((last_columnsize - targetColumnSize) / (float)1.1);
                else if (total > target) targetColumnSize -= Math.abs((last_columnsize - targetColumnSize) / (float)1.1);
                last_columnsize = save;
                last_total = total;
                first = false;
            }
            if (findMinimum) contentwidth = Math.round(total);
            else this.targetColumnSize = targetColumnSize;
        } while(false);
        //#end
    }

    void place() {
        solve(false);
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (!child.test(VISIBLE)) continue;
            if (!child.test(REPLACE)) continue;
            int child_width, child_height, child_x, child_y;
            if (!child.test(PACKED)) {
                child_width = child.test(HSHRINK) ? child.contentwidth : min(child.maxwidth, width - Math.abs(child.ax));
                child_height = child.test(VSHRINK) ? child.contentheight : min(child.maxheight, height - Math.abs(child.ay));
                child_width = max(child.minwidth, child_width);
                child_height = max(child.minheight, child_height);
                int gap_x = width - child_width;
                int gap_y = height - child_height;
                child_x = child.ax + (child.test(ALIGN_RIGHT) ? gap_x : !child.test(ALIGN_LEFT) ? gap_x / 2 : 0);
                child_y = child.ay + (child.test(ALIGN_BOTTOM) ? gap_y : !child.test(ALIGN_TOP) ? gap_y / 2 : 0);
            } else {
                int diff;
                //#repeat col/row colspan/rowspan contentwidth/contentheight width/height colMaxWidth/rowMaxHeight \
                //        child_x/child_y x/y HSHRINK/VSHRINK maxwidth/maxheight cols/rows minwidth/minheight x_slack/y_slack \
                //        child_width/child_height ALIGN_RIGHT/ALIGN_BOTTOM ALIGN_LEFT/ALIGN_TOP lp_h/lp \
                //        numregions/numregions_v regions/regions_v targetColumnSize/targetRowSize sizes/sizes_v
                child_x = 0;
                if (cols == 1) {
                    child_width = width;
                } else {
                    child_width = 0;
                    for(int r=0; r<numregions; r++) {
                        if (regions[r] < child.col) child_x += Math.round(sizes[r]);
                        else if (regions[r] < child.col+child.colspan) child_width += Math.round(sizes[r]);
                    }
                }
                diff = (child_width - (child.test(HSHRINK) ? child.contentwidth : min(child_width, child.maxwidth)));
                child_x += (child.test(ALIGN_RIGHT) ? diff : child.test(ALIGN_LEFT) ? 0 : diff / 2);
                child_width -= diff;
                //#end
            }
            if (test(MOVED)) child.set(MOVED);
            child.resize(child_x, child_y, child_width, child_height);
        }
        clear(MOVED);

        for(Box child = getChild(0); child != null; child = child.nextSibling())
            if (child.test(VISIBLE) && child.treeSize() > 0)
                child.place();
    }



    // Rendering Pipeline /////////////////////////////////////////////////////////////////////

    /** Renders self and children within the specified region. All rendering operations are clipped to xIn,yIn,wIn,hIn */
    public void render(int parentx, int parenty, int cx1, int cy1, int cx2, int cy2, PixelBuffer buf, Affine a) {
        if (!test(VISIBLE)) return;
        int globalx = parentx + (parent == null ? 0 : x);
        int globaly = parenty + (parent == null ? 0 : y);

        // intersect the x,y,w,h rendering window with ourselves; quit if it's empty
        if (test(CLIP)) {
            cx1 = max(cx1, globalx);
            cy1 = max(cy1, globaly);
            cx2 = min(cx2, globalx + width);
            cy2 = min(cy2, globaly + height);
            if (cx2 <= cx1 || cy2 <= cy1) return;
        }

        if ((fillcolor & 0xFF000000) != 0x00000000 || parent == null)
            buf.fillTrapezoid(cx1, cx2, cy1, cx1, cx2, cy2, (fillcolor & 0xFF000000) == 0 ? 0xffffffff : fillcolor);

        if (texture != null && texture.isLoaded)
            for(int x = globalx; x < cx2; x += texture.width)
                for(int y = globaly; y < cy2; y += texture.height)
                    buf.drawPicture(texture, x, y, cx1, cy1, cx2, cy2);
 
        if (text != null && !text.equals("") && font != null) {
            int gap_x = width - font.textwidth(text);
            int gap_y = height - font.textheight(text);
            int text_x = globalx + (test(ALIGN_RIGHT) ? gap_x : !test(ALIGN_LEFT) ? gap_x/2 : 0);
            int text_y = globaly + (test(ALIGN_BOTTOM) ? gap_y : !test(ALIGN_TOP) ? gap_y/2 : 0);
            font.rasterizeGlyphs(text, buf, strokecolor, text_x, text_y, cx1, cy1, cx2, cy2);
        }

        if (path != null) new Polygon(path, Affine.translate(globalx, globaly)).stroke(buf, strokecolor);

        for(Box b = getChild(0); b != null; b = b.nextSibling())
            b.render(globalx, globaly, cx1, cy1, cx2, cy2, buf, null);
    }
    
    
    // Methods to implement org.ibex.js.JS //////////////////////////////////////

  
    public JS call(JS method, JS[] args) throws JSExn {
        switch (args.length) {
            case 1: {
                //#switch(JSU.toString(method))
                case "indexof":
                    Box b = (Box)args[0];
                    if (b.parent != this)
                        return (redirect == null || redirect == this) ?
                            JSU.N(-1) : redirect.call(method, args);
                    return JSU.N(b.getIndexInParent());

                case "distanceto":
                    Box b = (Box)args[0];
                    JS ret = new JS.Obj();
                    ret.put(JSU.S("x"), JSU.N(b.localToGlobalX(0) - localToGlobalX(0)));
                    ret.put(JSU.S("y"), JSU.N(b.localToGlobalY(0) - localToGlobalY(0)));
                    return ret;

                //#end
            }
        }
        return super.call(method, args);
    }

    public JS get(JS name) throws JSExn {
        if (JSU.isInt(name))
            return redirect == null ? null : redirect == this ? getChild(JSU.toInt(name)) : redirect.get(name);

        //#switch(JSU.toString(name))
        case "surface": return parent == null ? null : parent.getAndTriggerTraps(name);
        case "indexof": return METHOD;
        case "distanceto": return METHOD;
        case "text": return JSU.S(text);
        case "path": {
            if (path != null) return JSU.S(path.toString());
            if (text == null) return null;
            if (font == null) return null;
            String ret = "";
            for(int i=0; i<text.length(); i++) ret += font.glyphs[text.charAt(i)].path;
            return JSU.S(ret);
        }
        case "fill": return JSU.S(Color.colorToString(fillcolor));
        case "strokecolor": return JSU.S(Color.colorToString(strokecolor));
        case "textcolor": return JSU.S(Color.colorToString(strokecolor));
        case "font": return font == null ? null : font.stream;
        case "fontsize": return font == null ? JSU.N(10) : JSU.N(font.pointsize);
        case "strokewidth": return JSU.N(strokewidth);
        case "align": return JSU.S(alignToString());
        case "thisbox": return this;
        case "shrink": return JSU.B(test(HSHRINK) || test(VSHRINK));
        case "hshrink": return JSU.B(test(HSHRINK));
        case "vshrink": return JSU.B(test(VSHRINK));
        case "aspect": return JSU.N(aspect);
        case "x": return (parent == null || !test(VISIBLE)) ? JSU.N(0) : test(PACKED) ? JSU.N(x) : JSU.N(ax);
        case "y": return (parent == null || !test(VISIBLE)) ? JSU.N(0) : test(PACKED) ? JSU.N(y) : JSU.N(ay);
        case "cols": return test(FIXED) == COLS ? JSU.N(cols) : JSU.N(0);
        case "rows": return test(FIXED) == ROWS ? JSU.N(rows) : JSU.N(0);
        case "colspan": return JSU.N(colspan);
        case "rowspan": return JSU.N(rowspan);
        case "width": getRoot().reflow(); return JSU.N(width);
        case "height": getRoot().reflow(); return JSU.N(height);
        case "minwidth": return JSU.N(minwidth);
        case "maxwidth": return JSU.N(maxwidth);
        case "minheight": return JSU.N(minheight);
        case "maxheight": return JSU.N(maxheight);
        case "clip": return JSU.B(test(CLIP));
        case "visible": return JSU.B(test(VISIBLE) && (parent == null || (parent.get(JSU.S("visible")) == JSU.T)));
        case "packed": return JSU.B(test(PACKED));
        case "globalx": return JSU.N(localToGlobalX(0));
        case "globaly": return JSU.N(localToGlobalY(0));
        case "cursor": return test(CURSOR) ? JSU.S((String)boxToCursor.get(this)) : null;
        case "mouse":
            if (getSurface() == null) return null;
            if (getSurface()._mousex == Integer.MAX_VALUE)
                throw new JSExn("you cannot read from the box.mouse property in background thread context");
            return new Mouse();
        case "numchildren": return redirect == null ? JSU.N(0) : redirect == this ? JSU.N(treeSize()) : redirect.get(JSU.S("numchildren"));
        case "redirect": return redirect == null ? null : redirect == this ? JSU.T : redirect.get(JSU.S("redirect"));
        case "Minimized": if (parent == null && getSurface() != null) return JSU.B(getSurface().minimized);
        default: return super.get(name);
        //#end
        throw new Error("unreachable"); // unreachable
    }

    private class Mouse extends JS.Immutable implements JS.Cloneable {
        public JS get(JS key) throws JSExn {
            //#switch(JSU.toString(key))
            case "x": return JSU.N(globalToLocalX(getSurface()._mousex));
            case "y": return JSU.N(globalToLocalY(getSurface()._mousey));

            // this might not get recomputed if we change mousex/mousey...
            case "inside": return JSU.B(test(MOUSEINSIDE));
            //#end
            return super.get(key);
        }
    }

    //#repeat setWidth/setHeight minwidth/minheight maxwidth/maxheight pendingWidth/pendingHeight
    public void setWidth(int min, int max) {
        // FIXME: deal with conflicting min/max
        if (this.minwidth == min && this.maxwidth == max) return;
        this.minwidth = min;
        this.maxwidth = max;
        RECONSTRAIN();
        if (parent != null || getSurface() == null) return;
        getSurface().pendingWidth = maxwidth;

        // FIXME: the repeat doesn't work right here
        getSurface().setMinimumSize(minwidth, minheight, minwidth != maxwidth || minheight != maxheight);
    }
    //#end

    public void put(JS name, JS value) throws JSExn {
        if (JSU.isInt(name)) { put(JSU.toInt(name), value); return; }
        //#switch(JSU.toString(name))
        case "thisbox":     if (value == null) removeSelf();
        case "text":        { String s = value == null ?  "" : JSU.toString(value); CHECKSET_STRING(text); RECONSTRAIN(); dirty(); }
        case "strokecolor": value = JSU.N(Color.stringToColor(JSU.toString(value))); CHECKSET_INT(strokecolor); dirty();
        case "textcolor":   value = JSU.N(Color.stringToColor(JSU.toString(value))); CHECKSET_INT(strokecolor); dirty();
        case "strokewidth": CHECKSET_SHORT(strokewidth); dirty();
        case "shrink":      CHECKSET_FLAG(HSHRINK | VSHRINK); RECONSTRAIN();
        case "hshrink":     CHECKSET_FLAG(HSHRINK); RECONSTRAIN();
        case "vshrink":     CHECKSET_FLAG(VSHRINK); RECONSTRAIN();
        case "path":        path = new Path(JSU.toString(value)); RECONSTRAIN(); dirty();
        case "width":       setWidth(JSU.toInt(value), JSU.toInt(value));
        case "height":      setHeight(JSU.toInt(value), JSU.toInt(value));
        case "maxwidth":    setWidth(minwidth, JSU.toInt(value));
        case "minwidth":    setWidth(JSU.toInt(value), maxwidth);
        case "maxheight":   setHeight(minheight, JSU.toInt(value));
        case "minheight":   setHeight(JSU.toInt(value), maxheight);
        case "colspan":     if (JSU.toInt(value) > 0) { CHECKSET_SHORT(colspan); if (parent != null) parent.REPACK(); }
        case "rowspan":     if (JSU.toInt(value) > 0) { CHECKSET_SHORT(rowspan); if (parent != null) parent.REPACK(); }
        case "visible":     CHECKSET_FLAG(VISIBLE); RECONSTRAIN(); dirty();
        case "packed":      CHECKSET_FLAG(PACKED); if (parent != null) { parent.REPACK(); } else { REPACK(); }
        case "align":       clear(ALIGNS); setAlign(value); REPLACE();
        case "cursor":      setCursor(JSU.toString(value));
        case "fill":        setFill(value);
        case "clip":        CHECKSET_FLAG(CLIP); if (parent == null) dirty(); else parent.dirty();
        case "rows": CHECKSET_SHORT(rows); if (rows==0){set(FIXED, COLS);if(cols==0)cols=1;} else set(FIXED, ROWS); REPACK();
        case "cols": CHECKSET_SHORT(cols); if (cols==0){set(FIXED, ROWS);if(rows==0)rows=1;} else set(FIXED, COLS); REPACK();

        // FIXME: remove
        case "mouse":
            int mousex = JSU.toInt(((JS)value).get(JSU.S("x")));
            int mousey = JSU.toInt(((JS)value).get(JSU.S("y")));
            getSurface()._mousex = localToGlobalX(mousex);
            getSurface()._mousey = localToGlobalY(mousey);

        case "Minimized": if (parent == null && getSurface() != null) getSurface().minimized = JSU.toBoolean(value);  // FEATURE
        case "Maximized": if (parent == null && getSurface() != null) getSurface().maximized = JSU.toBoolean(value);  // FEATURE
        case "Close":     if (parent == null && getSurface() != null) getSurface().dispose(true);
        case "redirect":
            for(Box cur = (Box)value; cur != null || cur == redirect; cur = cur.parent)
                if (cur == redirect) { redirect = (Box)value; return; }
            JSU.error("redirect can only be set to a descendant of its current value");
        case "fontsize": font = Font.getFont(font == null ? null : font.stream, JSU.toInt(value)); RECONSTRAIN(); dirty();
        case "font":
            if(!(value instanceof Stream)) throw new JSExn("You can only put streams to the font property");
            //FIXME: if (font == value) return;  // FIXME: unclone()
            font = value == null ? null : Font.getFont((Stream)value, font == null ? 10 : font.pointsize);
            RECONSTRAIN();
            dirty();
        case "x": if (parent==null && Surface.fromBox(this)!=null) {
            CHECKSET_INT(x);
        } else {
            if (test(PACKED) && parent != null) return;
            CHECKSET_INT(ax);
            REPLACE();
        }
        case "y": if (parent==null && Surface.fromBox(this)!=null) {
            CHECKSET_INT(y);
        } else {
            if (test(PACKED) && parent != null) return;
            CHECKSET_INT(ay);
            REPLACE();
        }
        case "titlebar": if (getSurface()!=null) getSurface().setTitleBarText(JSU.toString(value)); super.put(name,value);
        // FIXME: icon

        case "Press1":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Press2":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Press3":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release1":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release2":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release3":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click1":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click2":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click3":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick1":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick2":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick3":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "KeyPressed":    if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "KeyReleased":   if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Move":          if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "HScroll":       if (!test(STOP_UPWARD_PROPAGATION) && parent != null)
            parent.putAndTriggerTraps(name, JSU.N(JSU.toFloat(value) * ((float)parent.fontSize()) / ((float)fontSize())));
        case "VScroll":       if (!test(STOP_UPWARD_PROPAGATION) && parent != null)
            parent.putAndTriggerTraps(name, JSU.N(JSU.toFloat(value) * ((float)parent.fontSize()) / ((float)fontSize())));

        case "_Move":         propagateDownward(name, value, false);
        case "_Press1":       propagateDownward(name, value, false);
        case "_Press2":       propagateDownward(name, value, false);
        case "_Press3":       propagateDownward(name, value, false);
        case "_Release1":     propagateDownward(name, value, false);
        case "_Release2":     propagateDownward(name, value, false);
        case "_Release3":     propagateDownward(name, value, false);
        case "_Click1":       propagateDownward(name, value, false);
        case "_Click2":       propagateDownward(name, value, false);
        case "_Click3":       propagateDownward(name, value, false);
        case "_DoubleClick1": propagateDownward(name, value, false);
        case "_DoubleClick2": propagateDownward(name, value, false);
        case "_DoubleClick3": propagateDownward(name, value, false);
        case "_KeyPressed":   propagateDownward(name, value, false);
        case "_KeyReleased":  propagateDownward(name, value, false);
        case "_HScroll":      propagateDownward(name, value, false);
        case "_VScroll":      propagateDownward(name, value, false);

        case "SizeChange":    return;
        case "ChildChange":   return;
        case "Enter":         return;
        case "Leave":         return;

        default:              super.put(name, value);
        //#end
    }

    private String alignToString() {
        switch(flags & ALIGNS) {
            case (ALIGN_TOP | ALIGN_LEFT): return "topleft";
            case (ALIGN_BOTTOM | ALIGN_LEFT): return "bottomleft";
            case (ALIGN_TOP | ALIGN_RIGHT): return "topright";
            case (ALIGN_BOTTOM | ALIGN_RIGHT): return "bottomright";
            case ALIGN_TOP: return "top";
            case ALIGN_BOTTOM: return "bottom";
            case ALIGN_LEFT: return "left";
            case ALIGN_RIGHT: return "right";
            case 0: return "center";
            default: throw new Error("invalid alignment flags: " + (flags & ALIGNS));
        }
    }

    private void setAlign(JS value) throws JSExn {
        clear(ALIGNS);
        //#switch(JSU.toString(value))
        case "topleft": set(ALIGN_TOP | ALIGN_LEFT);
        case "bottomleft": set(ALIGN_BOTTOM | ALIGN_LEFT);
        case "topright": set(ALIGN_TOP | ALIGN_RIGHT);
        case "bottomright": set(ALIGN_BOTTOM | ALIGN_RIGHT);
        case "top": set(ALIGN_TOP);
        case "bottom": set(ALIGN_BOTTOM);
        case "left": set(ALIGN_LEFT);
        case "right": set(ALIGN_RIGHT);
        default: JSU.log("invalid alignment \"" + JSU.str(value) + "\"");
        //#end
    }
    
    private void setCursor(String value) throws JSExn {
        if (value == null) { clear(CURSOR); boxToCursor.remove(this); return; }
        if (value.equals(boxToCursor.get(this))) return;
        set(CURSOR);
        boxToCursor.put(this, value);
        Surface surface = getSurface();
        if (surface == null) return;
        String tempcursor = surface.cursor;
        propagateDownward(null, null, false);
        if (surface.cursor != tempcursor) surface.syncCursor();
    }

    private void setFill(JS value) throws JSExn {
        if (value == null) {
            if (texture == null && fillcolor == 0) return;
            texture = null;
            fillcolor = 0;
        } else if (JSU.isString(value)) {
            int newfillcolor = Color.stringToColor(JSU.toString(value));
            if (newfillcolor == fillcolor) return;
            fillcolor = newfillcolor;
            texture = null;
        } else {
            Picture newtex = Picture.load((JS)value, this);
            if (texture == newtex) return;
            texture = newtex;
            fillcolor = 0;
            if (texture != null && texture.isLoaded) run(null);
        }
        dirty();
    }

    /**
     *  Handles events which propagate down the box tree.  If obscured
     *  is set, then we merely check for Enter/Leave.
     */
    private void propagateDownward(JS name_, JS value, boolean obscured) throws JSExn {

        String name = JSU.toString(name_);
        if (getSurface() == null) return;
        int x = globalToLocalX(getSurface()._mousex);
        int y = globalToLocalY(getSurface()._mousey);
        boolean wasinside = test(MOUSEINSIDE);
        boolean isinside = test(VISIBLE) && inside(x, y) && !obscured;
        if (!wasinside && isinside) {
            set(MOUSEINSIDE);
            putAndTriggerTrapsAndCatchExceptions(JSU.S("Enter"), JSU.T);
        }
        if (isinside && test(CURSOR)) getSurface().cursor = (String)boxToCursor.get(this);
        if (wasinside && !isinside) {
            clear(MOUSEINSIDE);
            putAndTriggerTrapsAndCatchExceptions(JSU.S("Leave"), JSU.T);
        }

        boolean found = false;
        if (wasinside || isinside)
            for(Box child = getChild(treeSize() - 1); child != null; child = child.prevSibling()) {
                boolean save_stop = child.test(STOP_UPWARD_PROPAGATION);
                JS value2 = value;
                if (name.equals("_HScroll") || name.equals("_VScroll"))
                    value2 = JSU.N(JSU.toFloat(value) * ((float)child.fontSize()) / (float)fontSize());
                if (obscured || !child.inside(x - child.x, y - child.y)) {
                    child.propagateDownward(name_, value2, true);
                } else try {
                    found = true;
                    child.clear(STOP_UPWARD_PROPAGATION);
                    if (name != null) child.putAndTriggerTrapsAndCatchExceptions(name_, value2);
                    else child.propagateDownward(name_, value2, obscured);
                } finally {
                    if (save_stop) child.set(STOP_UPWARD_PROPAGATION); else child.clear(STOP_UPWARD_PROPAGATION);
                }
                if (child.inside(x - child.x, y - child.y))
                    if (name != null && name.equals("_Move")) obscured = true;
                    else break;
            }

        if (!obscured && !found)
            if ("_Move".equals(name) || name.startsWith("_Release") || wasinside)
                if (name != null)
                    putAndTriggerTrapsAndCatchExceptions(JSU.S(name.substring(1)), value);
    }

    /** figures out what box in this subtree of the Box owns the pixel at x,y relitave to the Surface */
    public static Box whoIs(Box cur, int x, int y) {

        if (cur.parent != null) throw new Error("whoIs may only be invoked on the root box of a surface");
        int globalx = 0;
        int globaly = 0;

        // WARNING: this method is called from the event-queueing thread -- it may run concurrently with
        // ANY part of Ibex, and is UNSYNCHRONIZED for performance reasons.  BE CAREFUL HERE.

        if (!cur.test(VISIBLE)) return null;
        if (!cur.inside(x - globalx, y - globaly)) return cur.parent == null ? cur : null;
        OUTER: while(true) {
            for(int i=cur.treeSize() - 1; i>=0; i--) {
                Box child = cur.getChild(i);
                if (child == null) continue;        // since this method is unsynchronized, we have to double-check
                globalx += child.x;
                globaly += child.y;
                if (child.test(VISIBLE) && child.inside(x - globalx, y - globaly)) { cur = child; continue OUTER; }
                globalx -= child.x;
                globaly -= child.y;
            }
            break;
        }
        return cur;
    }


    // Trivial Helper Methods (should be inlined) /////////////////////////////////////////

    void mark_for_repack() { REPACK(); }
    public Enumeration keys() { throw new Error("you cannot apply for..in to a " + this.getClass().getName()); }
    public Box getRoot() { return parent == null ? this : parent.getRoot(); }
    public Surface getSurface() { return Surface.fromBox(getRoot()); }
    Box nextPackedSibling() { Box b = nextSibling(); return b == null || (b.test(PACKED | VISIBLE)) ? b : b.nextPackedSibling(); }
    Box firstPackedChild() { Box b = getChild(0); return b == null || (b.test(PACKED | VISIBLE)) ? b : b.nextPackedSibling(); }
    public int globalToLocalX(int x) { return parent == null ? x : parent.globalToLocalX(x - this.x); }
    public int globalToLocalY(int y) { return parent == null ? y : parent.globalToLocalY(y - this.y); }
    public int localToGlobalX(int x) { return parent == null ? x : parent.globalToLocalX(x + this.x); }
    public int localToGlobalY(int y) { return parent == null ? y : parent.globalToLocalY(y + this.y); }

    static short min(short a, short b) { if (a<b) return a; else return b; }
    static int min(int a, int b) { if (a<b) return a; else return b; }
    static float min(float a, float b) { if (a<b) return a; else return b; }

    static short max(short a, short b) { if (a>b) return a; else return b; }
    static int max(int a, int b) { if (a>b) return a; else return b; }
    static float max(float a, float b) { if (a>b) return a; else return b; }

    static int min(int a, int b, int c) { if (a<=b && a<=c) return a; else if (b<=c && b<=a) return b; else return c; }
    static int max(int a, int b, int c) { if (a>=b && a>=c) return a; else if (b>=c && b>=a) return b; else return c; }
    static int bound(int a, int b, int c) { if (c < b) return c; if (a > b) return a; return b; }
    final boolean inside(int x, int y) { return test(VISIBLE) && x >= 0 && y >= 0 && x < width && y < height; }

    void set(int mask) { flags |= mask; }
    void set(int mask, boolean setclear) { if (setclear) set(mask); else clear(mask); }
    void clear(int mask) { flags &= ~mask; }
    public boolean test(int mask) { return ((flags & mask) == mask); }
    

    // Tree Handling //////////////////////////////////////////////////////////////////////

    public final int getIndexInParent() { return parent == null ? 0 : parent.indexNode(this); }
    public final Box nextSibling() { return parent == null ? null : parent.getChild(parent.indexNode(this) + 1); }
    public final Box prevSibling() { return parent == null ? null : parent.getChild(parent.indexNode(this) - 1); }
    public final Box getChild(int i) {
        if (i < 0) return null;
        if (i >= treeSize()) return null;
        return (Box)getNode(i);
    }

    // Tree Manipulation /////////////////////////////////////////////////////////////////////

    public void removeSelf() {
        if (parent != null) { parent.removeChild(parent.indexNode(this)); return; }
        Surface surface = Surface.fromBox(this); 
        if (surface != null) surface.dispose(true);
    }

    /** remove the i^th child */
    public void removeChild(int i) {
        Box b = getChild(i);
        b.RECONSTRAIN();
        b.dirty();
        b.clear(MOUSEINSIDE);
        deleteNode(i);
        b.parent = null;
        REPACK();
        putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);
    }
    
    public void put(int i, JS value) throws JSExn {
        if (i < 0) return;
            
        if (value != null && !(value instanceof Box)) {
            if (Log.on) JSU.warn("attempt to set a numerical property on a box to a non-box");
            return;
        }

        if (redirect == null) {
            if (value == null) putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), getChild(i));
            else JSU.warn("attempt to add/remove children to/from a node with a null redirect");

        } else if (redirect != this) {
            if (value != null) putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), value);
            redirect.put(i, value);
            if (value == null) {
                Box b = (Box)redirect.get(JSU.N(i));
                if (b != null) putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);
            }

        } else if (value == null) {
            if (i < 0 || i > treeSize()) return;
            Box b = getChild(i);
            removeChild(i);
            putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);

        } else {
            Box b = (Box)value;

            // check if box being moved is currently target of a redirect
            for(Box cur = b.parent; cur != null; cur = cur.parent)
                if (cur.redirect == b) {
                    if (Log.on) JSU.warn("attempt to move a box that is the target of a redirect");
                    return;
                }

            // check for recursive ancestor violation
            for(Box cur = this; cur != null; cur = cur.parent)
                if (cur == b) {
                    if (Log.on) JSU.warn("attempt to make a node a parent of its own ancestor");
                    if (Log.on) Log.info(this, "box == " + this + "  ancestor == " + b);
                    return;
                }

            if (b.parent != null) b.parent.removeChild(b.parent.indexNode(b));
            insertNode(i, b);
            b.parent = this;
            
            // need both of these in case child was already uncalc'ed
            b.REPACK();
            REPACK();
            
            b.dirty(); 
            putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);
        }
    }
    
    public void putAndTriggerTrapsAndCatchExceptions(JS name, JS val) {
        try {
            putAndTriggerTraps(name, val);
        } catch (JSExn e) {
            JSU.log("caught js exception while putting to trap \""+ JSU.str(name)+"\"");
            JSU.log(e);
        } catch (Exception e) {
            JSU.log("caught exception while putting to trap \""+ JSU.str(name)+"\"");
            JSU.log(e);
        }
    }
    
    // BalancedTree functions
    private void insertNode(int p, Box b) {
        if(bt == null) bt = new BalancedTree();
        bt.insertNode(p,b);
    }
    
    private int treeSize() { return bt == null ? 0 : bt.treeSize(); }
    private int indexNode(Box b) { return bt == null ? -1 : bt.indexNode(b); }
    private void deleteNode(int p) { bt.deleteNode(p); }
    private Box getNode(int p) { return (Box)bt.getNode(p); }
}
